
import math
import numpy as np

import large_image
from large_image.cache_util import LruCacheMetaclass, methodcache, strhash

import OpenVisus as ov

# ///////////////////////////////////////////////////////////////////////////////////
class OpenVisusTileSource(large_image.tilesource.TileSource, metaclass=LruCacheMetaclass):

	cacheName = 'tilesource'
	name = 'openvisus'
 
	min_tile_pixels =256*256

	extensions = {
	  None: large_image.constants.SourcePriority.LOW,
	  'idx': large_image.constants.SourcePriority.PREFERRED,
	}

	# constructor
	def __init__(self, url=None, **kwargs):

		super().__init__(**kwargs)
  
		self.db=ov.LoadDataset(url)
		self.bitmask=self.db.getBitmask().toString()
		print(self.bitmask)
		self.sizeY,self.sizeX=self.db.shape

		# TODO?
		assert('2' not in self.bitmask) 
		self.levels=1
		self.tileWidth =2**len([it for it in self.bitmask if it=='0'])
		self.tileHeight=2**len([it for it in self.bitmask if it=='1'])
		print("pow2 size",self.tileWidth,self.tileHeight)
		
		# the right part must be 01 or 10 this to keep the tile size constant
		s=self.bitmask
		while (s.endswith("01") or s.endswith("10")):
    
			# do not go too small
			if self.tileWidth*self.tileHeight==self.min_tile_pixels:
				break
  
			self.tileWidth >>=1
			self.tileHeight>>=1
			self.levels+=1
			s=s[0:-2]
  
		self.minLevel = 0
		self.maxLevel = self.levels-1
  
		#self.dtype=self.db.getField().dtype
		#self.bandCount=1
  
		print(f"OpenVisusTileSource({url},kwargs={kwargs})")
  

		# TODO
		# fields? how to support multiple fields?
		# bands?  I think I just return the image
		# frames? could be 3d or with time, I do not think both
		#         if 3d then I need to make sure the Z samples are all on the left, like V22220101010101....
		
	# canRead
	@classmethod
	def canRead(cls, *args, **kwargs):
		return True

	# getMetadata
	def getMetadata(self):
		return super().getMetadata()

	# getTile
	@methodcache()
	def getTile(self, x, y, z, *args, **kwargs):
   
		# I am always returning numpy arrays
		assert(kwargs.get("numpyAllowed","always")=="always")
  
		# todo
		assert(kwargs.get("frame",None) is None)

		max_resolution=len(self.bitmask)-1
		dx=self.tileWidth
		dy=self.tileHeight
		for L in range(self.maxLevel,z,-1):
			dx<<=1
			dy<<=1
			max_resolution-=2
	
		x1 = x*dx
		y1 = y*dy
		x2 = min(x1+dx,self.sizeX)
		y2 = min(y1+dy,self.sizeY)
  
		# mirror y ?
	
		print(f"getTile x={x} y={y} z={z} tileWidth={self.tileWidth} tileHeight={self.tileHeight} x1={x1} y1={y1} x2={x2} y2={y2} max_resolution={max_resolution}")
		tile=self.db.read(x=[x1,x2],y=[y1,y2],max_resolution=max_resolution)
		# can be smaller because I am on the boundary
		assert(tile.shape[0]<=self.tileHeight and tile.shape[1]<=self.tileWidth)

		return self._outputTile(tile, large_image.constants.TILE_FORMAT_NUMPY, x, y, z, **kwargs)

	@staticmethod
	def getLRUHash(*args, **kwargs):
		return strhash(super(OpenVisusTileSource, OpenVisusTileSource).getLRUHash(*args, **kwargs))

# //////////////////////////////////////////////////////////////
def open(*args, **kwargs):
	return OpenVisusTileSource(*args, **kwargs)

# //////////////////////////////////////////////////////////////
def canRead(*args, **kwargs):
	return OpenVisusTileSource.canRead(*args, **kwargs)

